Require Import ssreflect ssrfun eqtype ssrbool ssrnat seq choice finset div
  fintype tuple finfun bigop prime binomial ssralg sdetc.

Require Import ZArith.

(* In this file, we test the determinant function as described and proved
 correct in the file sdetc. We first Show that Z can be endowed with a
 commutative ring structure. *)

Lemma Zeq_axiom : Equality.axiom Zeq_bool.
rewrite /Equality.axiom => x y; case h: (Zeq_bool x y).
  by apply: ReflectT; apply/Zeq_is_eq_bool. 
by apply ReflectF; apply/Zeq_is_eq_bool; rewrite h.
Qed.

Definition Zpickle (x : Z) : nat :=
  match x with
    Z0 => 0%nat
  | Zpos p => ((nat_of_P p).*2 - 1)%nat
  | Zneg p => ((nat_of_P p).*2)%nat
  end.

Definition Zunpickle (x : nat) : Z :=
  if x == 0 then Z0 else
  if odd x then Zpos (P_of_succ_nat (half x)) else
   Zneg (P_of_succ_nat (half (x - 1))).

Lemma SZ_pickleK : cancel Zpickle Zunpickle.
move => [ | p | p]; first done; rewrite /Zpickle /Zunpickle;
 move/ltP: (lt_O_nat_of_P p); case pk : (nat_of_P p) => [ | k] _ //.
 have -> : (k.+1.*2 - 1 == 0) = false by rewrite doubleS subSS subn0.
 have -> : odd (k.+1.*2 - 1) by rewrite odd_sub // odd_double.
 have -> : k.+1.*2 -1 = 1 + k.*2 by rewrite doubleS subSS subn0 addSn add0n.
 rewrite (half_bit_double _ true); apply f_equal.
 by apply: nat_of_P_inj; rewrite nat_of_P_o_P_of_succ_nat_eq_succ.
rewrite (_ : (k.+1.*2 == 0) = false) // (_ : k.+1.*2 - 1 = 1 + k.*2) //.
rewrite odd_double (half_bit_double _ true); apply f_equal.
by apply: nat_of_P_inj; rewrite nat_of_P_o_P_of_succ_nat_eq_succ.
Qed.

Lemma n10Z : ~~Zeq_bool 1 0.
Proof. by []. Qed.

Definition SZ_eqMixin :=  Equality.Mixin Zeq_axiom.
Canonical Structure SZ_eqType : eqType := Equality.pack SZ_eqMixin.
Definition SZ_countable_mixin := CanCountMixin SZ_pickleK.
Definition SZ_choiceMixin := CountChoiceMixin SZ_countable_mixin.
Canonical Structure SZ_choiceType := Eval hnf in ChoiceType Z SZ_choiceMixin.
Canonical Structure SZ_countType := Eval hnf in CountType Z SZ_countable_mixin.
Definition SZ_ZmodMixin :=
  ZmodMixin Zplus_assoc Zplus_comm Zplus_0_l Zplus_opp_l.
Canonical Structure SZ_ZmodType := ZmodType Z SZ_ZmodMixin.
Definition SZ_comRingMixin : GRing.Ring.mixin_of SZ_ZmodType := 
  ComRingMixin Zmult_assoc Zmult_comm Zmult_1_l Zmult_plus_distr_l n10Z.
Canonical Structure SZ_Ring := Eval hnf in RingType Z SZ_comRingMixin.
Canonical Structure SZ_comRing := Eval hnf in ComRingType Z Zmult_comm.

Eval vm_compute in sdet 2 ([:: [:: 1; 2]; [::3; 4]])%Z.

Eval vm_compute in sdet 4
 ([::[::  1;  2;  3;  4];
     [::  5;  6;  7;  8];
     [::  9; 10; 28;  0];
     [::  3; 2;   2; -1]])%Z.
